import json
from typing import Tuple, List

from sklearn.model_selection import train_test_split

from headliner.model.summarizer_transformer import SummarizerTransformer
from headliner.trainer import Trainer


def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':
    data_raw = read_data('/Users/cschaefe/datasets/en_ger.txt')[:10000]
    train_data, val_data = train_test_split(data_raw, test_size=512, shuffle=True, random_state=42)

    summarizer = SummarizerTransformer(num_heads=1,
                                       feed_forward_dim=1024,
                                       embedding_size=50,
                                       embedding_encoder_trainable=True,
                                       embedding_decoder_trainable=True,
                                       dropout_rate=0,
                                       max_prediction_len=20)

    trainer = Trainer(steps_per_epoch=500,
                      batch_size=32,
                      steps_to_log=5,
                     # embedding_path_encoder='/Users/cschaefe/workspace/GloVe/glove_welt_dedup_128.txt',
                     # embedding_path_decoder='/Users/cschaefe/workspace/GloVe/glove_welt_dedup_128.txt',
                      model_save_path=None,
                      tensorboard_dir='/tmp/nobucketing',
                      bucketing_batches_to_bucket=1)

    trainer.train(summarizer, train_data, val_data=val_data)





import unittest

import numpy as np

from headliner.preprocessing.dataset_generator import DatasetGenerator


class TestDatasetGenerator(unittest.TestCase):

    def test_generate_dataset(self):
        data = [([1, 1], [2, 2]), ([1, 1, 1], [2, 2])]
        batch_generator = DatasetGenerator(batch_size=1)

        # batch size = 1
        batches_iter = iter(batch_generator(lambda: data))
        batches = next(batches_iter)
        print(batches[0].numpy().tolist())
        expected = [[[1, 1]], [[2, 2]]]
        np.testing.assert_array_equal(expected[0], batches[0].numpy().tolist())
        np.testing.assert_array_equal(expected[1], batches[1].numpy().tolist())

        # batch size = 2
        batch_generator = DatasetGenerator(batch_size=2)
        batches_iter = iter(batch_generator(lambda: data))
        batches = next(batches_iter)
        expected = [[[1, 1, 0], [1, 1, 1]], [[2, 2], [2, 2]]]
        np.testing.assert_array_equal(expected[0], batches[0].numpy().tolist())
        np.testing.assert_array_equal(expected[1], batches[1].numpy().tolist())



from typing import List, Iterable, Dict

from keras_preprocessing.text import Tokenizer as KTokenizer

from headliner.preprocessing.tokenizer import Tokenizer


class KerasTokenizer(Tokenizer):

    def __init__(self, **kwargs):
        self._keras_tokenizer = KTokenizer(**kwargs)

    def encode(self, text: str) -> List[int]:
        return self._keras_tokenizer.texts_to_sequences([text])[0]

    def decode(self, sequence: List[int]) -> str:
        return self._keras_tokenizer.sequences_to_texts([sequence])[0]

    @property
    def vocab_size(self) -> int:
        return len(self._keras_tokenizer.word_index)

    def fit(self, texts: Iterable[str]):
        self._keras_tokenizer.fit_on_texts(texts)

    @property
    def token_index(self) -> Dict[str, int]:
        return self._keras_tokenizer.word_index



from typing import Dict

import numpy as np


def read_embedding(file_path: str, vector_dim: int) -> Dict[str, np.array]:
    """
    Reads an embedding file in glove format into a dictionary mapping tokens to vectors.
    """

    glove = {}
    with open(file_path, encoding='utf-8') as f:
        for line in f:
            values = line.split()
            # handle weird whitespaces in tokens
            if len(values[1:]) > vector_dim:
                wordcount = len(values) - vector_dim
                vec = np.asarray(values[wordcount:], dtype='float32')
            else:
                vec = np.asarray(values[1:], dtype='float32')
            token = values[0]
            glove[token] = vec
    return glove


def embedding_to_matrix(embedding: Dict[str, np.array],
                        token_index: Dict[str, int],
                        embedding_dim: int) -> np.array:
    """
    Converts an embedding dictionary into a weights matrix used to initialize an embedding layer.
    It ensures that all tokens in the token_index dictionare are mapped to a row, even those that are
    not contained in the provided embedding dictionary. Unknown tokens are initialized with a random
    vector with entries between -1 and 1.

    Args:
        embedding: dictionary mapping tokens to embedding vectors
        token_index: dictionary mapping tokens to indices that are fed into the embedding layer
        embedding_dim: size of the embedding vectors

    Returns: embedding weights as numpy array
    """
    np.random.seed(42)
    embedding_matrix = 2. * np.random.rand(len(token_index) + 1, embedding_dim) - 1.
    for token, index in token_index.items():
        embedding_vec = embedding.get(token)
        if embedding_vec is not None:
            embedding_matrix[index] = embedding_vec
    return embedding_matrix



from headliner.model.summarizer_transformer import SummarizerTransformer

if __name__ == '__main__':
    path_to_model = '/tmp/summarizer_20191010_120007'
    summarizer = SummarizerTransformer.load(path_to_model)

    while True:
        text = input('\nEnter text: ')
        prediction_vecs = summarizer.predict_vectors(text, '')
        tokens_input = prediction_vecs['preprocessed_text'][0].split()
        print('\n')
        print(prediction_vecs['predicted_text'])
        print('\n')
        """
        alignment = prediction_vecs['alignment']
        for t in range(len(alignment)):
            al = alignment[t]
            top_att = np.argsort(al)[::-1]
            top_ind = top_att.tolist()
            top_tokens = [tokens_input[i] for i in top_ind][:5]
            print('{} {}'.format(prediction_vecs['predicted_text'].split()[t], top_tokens))
        sum_alignment = sum(alignment)
        top_att = np.argsort(sum_alignment)[::-1]
        top_ind = top_att.tolist()
        top_tokens = [tokens_input[i] for i in top_ind][:10]
        print()
        print(top_tokens)
        print()
        """


from setuptools import setup, find_packages

long_description = '''
Headliner is a library that was internally created to generate headlines from news articles. 
In particular, we use sequence-to-sequence (seq2seq) under the hood, 
an encoder-decoder framework. We provide a very simple interface to train 
and deploy seq2seq models. Although this library was created internally to 
generate headlines, you can also use it for other tasks like machine translations,
text summarization and many more.

Read the documentation at: https://as-ideas.github.io/headliner/

Headliner is compatible with Python 3.6 and is distributed under the MIT license.
'''

setup(
    name='headliner',
    version='0.0.15',
    author='Christian Schäfer',
    author_email='c.schaefer.home@gmail.com',
    description='Generating headlines from news articles using seq2seq models.',
    long_description=long_description,
    license='MIT',
    install_requires=['scikit-learn', 'nltk', 'pyyaml'],
    extras_require={
        'tests': ['pytest', 'pytest-cov', 'codecov', 'tensorflow~=2.0.0'],
        'docs': ['mkdocs', 'mkdocs-material'],
        'dev': ['bumpversion']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=('tests',)),
)


import abc
from abc import abstractmethod
from typing import Callable, Dict, Union

import numpy as np
import tensorflow as tf

from headliner.preprocessing import Preprocessor, Vectorizer


class Summarizer(abc.ABC):

    def __init__(self):
        self.vectorizer: Union[Vectorizer, None] = None
        self.preprocessor: Union[Preprocessor, None] = None
        self.embedding_size: Union[int, None] = None

    @abstractmethod
    def init_model(self,
                   preprocessor: Preprocessor,
                   vectorizer: Vectorizer,
                   embedding_weights_encoder=None,
                   embedding_weights_decoder=None) -> None:
        """
        Initializes the model and provides necessary information for compilation.

        Args:
            preprocessor: Preprocessor object that preprocesses text for training and prediction.
            vectorizer: Vectorizer object that performs vectorization of the text.
            embedding_weights_encoder (optional): Matrix to initialize the encoder embedding.
            embedding_weights_decoder (optional): Matrix to initialize the decoder embedding.
        """

        pass

    @abstractmethod
    def predict(self, text: str) -> str:
        """
        Predicts summary of an input text.
        """

        pass

    @abstractmethod
    def predict_vectors(self, input_text: str, target_text: str) -> Dict[str, Union[str, np.array]]:
        """
        Predicts summary of an input text and outputs information needed for evaluation:
        output logits, input tokens, output tokens, predicted tokens, preprocessed text,
        attention alignment.

        Args:
            input_text: Text used as input for prediction.
            target_text: Text used for evaluation.

        Returns: Dictionary with prediction information such as
            preprocessed_text, logits, alignment, predicted_sequence, predicted_text.

        """

        pass

    @abstractmethod
    def new_train_step(self,
                       loss_function: Callable[[tf.Tensor], tf.Tensor],
                       batch_size: int,
                       apply_gradients=True) -> Callable[[tf.Tensor, tf.Tensor], float]:

        """
        Initializes the train_step function to train the model on batches of data.

        Args:
            loss_function: Loss function to perform backprop on.
            batch_size: Batch size to use for training.
            apply_gradients: Whether to apply the gradients, i.e.
                False if you want to validate the model on test data.

        Returns: Train step function that is applied to a batch and returns the loss.

        """

        pass

    @abstractmethod
    def save(self, out_path: str) -> None:
        """
        Saves the model to a file.

        Args:
            out_path: Path to directory for saving the model.

        """

        pass

    @staticmethod
    @abstractmethod
    def load(in_path: str):
        """
        Loads the model from a file.

        Args:
            in_path: Path to the model directory.

        Returns: Instance of the loaded summarizer.

        """

        pass



import unittest
from headliner.preprocessing.keras_tokenizer import KerasTokenizer


class TestKerasTokenizer(unittest.TestCase):

    def test_keras_tokenizer(self):
        tokenizer = KerasTokenizer(filters='', lower=False, oov_token='<unk>')
        tokenizer.fit(['a b c d'])
        encoded = tokenizer.encode('a b e')
        self.assertEqual([2, 3, 1], encoded)
        decoded = tokenizer.decode(encoded)
        self.assertEqual('a b <unk>', decoded)
        self.assertEqual(5, tokenizer.vocab_size)
        self.assertEqual({'a', 'b', 'c', 'd', '<unk>'}, tokenizer.token_index.keys())
        self.assertEqual({1, 2, 3, 4, 5}, set(tokenizer.token_index.values()))




import datetime
import logging
from collections import Counter
from typing import Tuple, List, Iterable, Callable, Dict, Union

import tensorflow as tf
import yaml

from headliner.callbacks.evaluation_callback import EvaluationCallback
from headliner.callbacks.model_checkpoint_callback import ModelCheckpointCallback
from headliner.callbacks.validation_callback import ValidationCallback
from headliner.embeddings import read_embedding, embedding_to_matrix
from headliner.evaluation.scorer import Scorer
from headliner.losses import masked_crossentropy
from headliner.model.summarizer import Summarizer
from headliner.preprocessing.bucket_generator import BucketGenerator
from headliner.preprocessing.dataset_generator import DatasetGenerator
from headliner.preprocessing.keras_tokenizer import KerasTokenizer
from headliner.preprocessing.preprocessor import Preprocessor
from headliner.preprocessing.vectorizer import Vectorizer
from headliner.utils.logger import get_logger

START_TOKEN = '<start>'
END_TOKEN = '<end>'
OOV_TOKEN = '<unk>'


class Trainer:

    def __init__(self,
                 max_output_len=None,
                 batch_size=16,
                 max_vocab_size_encoder=200000,
                 max_vocab_size_decoder=200000,
                 embedding_path_encoder=None,
                 embedding_path_decoder=None,
                 steps_per_epoch=500,
                 tensorboard_dir='/tmp/train_tens_' + datetime.datetime.now().strftime('%Y%m%d_%H%M%S'),
                 model_save_path='/tmp/summarizer_' + datetime.datetime.now().strftime('%Y%m%d_%H%M%S'),
                 bucketing_buffer_size_batches=10000,
                 bucketing_batches_to_bucket=100,
                 logging_level=logging.INFO,
                 num_print_predictions=5,
                 steps_to_log=10,
                 preprocessor: Union[Preprocessor, None] = None) -> None:
        """
        Initializes the trainer.

        Args:
            max_output_len: Maximum length of output sequences. Default None, then eager mode will be enabled,
                else static.
            batch_size: Size of mini-batches for stochastic gradient descent.
            max_vocab_size_encoder: Maximum number of unique tokens to consider for encoder embeddings.
            max_vocab_size_decoder: Maximum number of unique tokens to consider for decoder embeddings.
            embedding_path_encoder: Path to embedding file for the encoder.
            embedding_path_decoder: Path to embedding file for the decoder.
            steps_per_epoch: Number of steps to train until callbacks are invoked.
            tensorboard_dir: Directory for saving tensorboard logs.
            model_save_path: Directory for saving the best model.
            bucketing_buffer_size_batches: Number of batches to buffer when bucketing sequences.
            bucketing_batches_to_bucket: Number of buffered batches from which sequences are collected for bucketing.
            logging_level: Level of logging to use, e.g. logging.INFO or logging.DEBUG.
            num_print_predictions: Number of sample predictions to print in each evaluation.
            steps_to_log: Number of steps to wait for logging output.
            preprocessor (optional): custom preprocessor, if None a standard preprocessor will be created.
        """

        self.max_output_len = max_output_len
        self.batch_size = batch_size
        self.max_vocab_size_encoder = max_vocab_size_encoder
        self.max_vocab_size_decoder = max_vocab_size_decoder
        self.bucketing_buffer_size_batches = bucketing_buffer_size_batches
        self.bucketing_batches_to_bucket = bucketing_batches_to_bucket
        self.embedding_path_encoder = embedding_path_encoder
        self.embedding_path_decoder = embedding_path_decoder
        self.steps_per_epoch = steps_per_epoch
        self.tensorboard_dir = tensorboard_dir
        self.model_save_path = model_save_path
        self.loss_function = masked_crossentropy
        self.dataset_generator = DatasetGenerator(self.batch_size)
        self.bucket_generator = BucketGenerator(element_length_function=lambda vecs: len(vecs[0]),
                                                batch_size=self.batch_size,
                                                buffer_size_batches=self.bucketing_buffer_size_batches,
                                                batches_to_bucket=self.bucketing_batches_to_bucket,
                                                shuffle=True,
                                                seed=42)
        self.logger = get_logger(__name__)
        self.logger.setLevel(logging_level)
        self.num_print_predictions = num_print_predictions
        self.steps_to_log = steps_to_log
        self.preprocessor = preprocessor or Preprocessor(start_token=START_TOKEN, end_token=END_TOKEN)

    @classmethod
    def from_config(cls, file_path, **kwargs):
        with open(file_path, 'r', encoding='utf-8') as f:
            cfg = yaml.load(f)
            batch_size = cfg['batch_size']
            max_vocab_size_encoder = cfg['max_vocab_size_encoder']
            max_vocab_size_decoder = cfg['max_vocab_size_decoder']
            glove_path_encoder = cfg['embedding_path_encoder']
            glove_path_decoder = cfg['embedding_path_decoder']
            steps_per_epoch = cfg['steps_per_epoch']
            tensorboard_dir = cfg['tensorboard_dir']
            model_save_path = cfg['model_save_path']
            bucketing_buffer_size_batches = cfg['bucketing_buffer_size_batches']
            bucketing_batches_to_bucket = cfg['bucketing_batches_to_bucket']
            steps_to_log = cfg['steps_to_log']
            logging_level = logging.INFO
            logging_level_string = cfg['logging_level']
            max_output_len = cfg['max_output_len']
            if logging_level_string == 'debug':
                logging_level = logging.DEBUG
            elif logging_level_string == 'error':
                logging_level = logging.ERROR
            return Trainer(batch_size=batch_size,
                           max_vocab_size_encoder=max_vocab_size_encoder,
                           max_vocab_size_decoder=max_vocab_size_decoder,
                           embedding_path_encoder=glove_path_encoder,
                           embedding_path_decoder=glove_path_decoder,
                           steps_per_epoch=steps_per_epoch,
                           tensorboard_dir=tensorboard_dir,
                           model_save_path=model_save_path,
                           bucketing_buffer_size_batches=bucketing_buffer_size_batches,
                           bucketing_batches_to_bucket=bucketing_batches_to_bucket,
                           logging_level=logging_level,
                           steps_to_log=steps_to_log,
                           max_output_len=max_output_len,
                           **kwargs)

    def train(self,
              summarizer: Summarizer,
              train_data: Iterable[Tuple[str, str]],
              val_data: Iterable[Tuple[str, str]] = None,
              num_epochs=2500,
              scorers: Dict[str, Scorer] = None,
              callbacks: List[tf.keras.callbacks.Callback] = None) -> None:
        """
        Trains a summarizer or resumes training of a previously initialized summarizer.

        Args:
            summarizer: Model to train, can be either a freshly created model or a loaded model.
            train_data: Data to train the model on.
            val_data (optional): Validation data.
            num_epochs: Number of epochs to train.
            scorers (optional): Dictionary with {score_name, scorer} to add validation scores to the logs.
            callbacks (optional): Additional custom callbacks.
        """
        if summarizer.preprocessor is None or summarizer.vectorizer is None:
            self.logger.info('training a bare model, preprocessing data to init model...')
            self._init_model(summarizer, train_data)
        else:
            self.logger.info('training an already initialized model...')
        vectorize_train = self._vectorize_data(preprocessor=summarizer.preprocessor,
                                               vectorizer=summarizer.vectorizer,
                                               bucket_generator=self.bucket_generator)
        vectorize_val = self._vectorize_data(preprocessor=summarizer.preprocessor,
                                             vectorizer=summarizer.vectorizer,
                                             bucket_generator=None)
        train_dataset = self.dataset_generator(lambda: vectorize_train(train_data))
        val_dataset = self.dataset_generator(lambda: vectorize_val(val_data))

        train_callbacks = callbacks or []
        if val_data is not None:
            train_callbacks.extend([
                EvaluationCallback(summarizer=summarizer,
                                   scorers=scorers or {},
                                   val_data=val_data,
                                   print_num_examples=self.num_print_predictions),
                ValidationCallback(summarizer=summarizer,
                                   val_dataset=val_dataset,
                                   loss_function=self.loss_function,
                                   batch_size=self.batch_size),
                ModelCheckpointCallback(file_path=self.model_save_path,
                                        summarizer=summarizer,
                                        monitor='loss_val',
                                        mode='min')
            ])
        if self.tensorboard_dir is not None:
            train_callbacks.append(tf.keras.callbacks.TensorBoard(log_dir=self.tensorboard_dir,
                                                                  update_freq='epoch'))
        logs = {}
        epoch_count, batch_count = 0, 0
        train_step = summarizer.new_train_step(self.loss_function, self.batch_size, apply_gradients=True)
        while epoch_count < num_epochs:
            for train_source_seq, train_target_seq in train_dataset.take(-1):
                batch_count += 1
                current_loss = float(train_step(train_source_seq, train_target_seq))
                logs['loss'] = current_loss
                if batch_count % self.steps_to_log == 0:
                    self.logger.info('epoch {epoch}, batch {batch}, logs: {logs}'.format(epoch=epoch_count,
                                                                                         batch=batch_count,
                                                                                         logs=logs))
                if batch_count % self.steps_per_epoch == 0:
                    for callback in train_callbacks:
                        callback.on_epoch_end(epoch_count, logs=logs)
                    epoch_count += 1
                    if epoch_count >= num_epochs:
                        break

            self.logger.info('finished iterating over dataset, total batches: {}'.format(batch_count))
            if batch_count == 0:
                raise ValueError('Iterating over the dataset yielded zero batches!')

    def _init_model(self,
                    summarizer: Summarizer,
                    train_data: Iterable[Tuple[str, str]]) -> None:

        tokenizer_encoder, tokenizer_decoder = self._create_tokenizers(train_data)
        self.logger.info('vocab encoder: {vocab_enc}, vocab decoder: {vocab_dec}'.format(
            vocab_enc=tokenizer_encoder.vocab_size, vocab_dec=tokenizer_decoder.vocab_size))
        vectorizer = Vectorizer(tokenizer_encoder,
                                tokenizer_decoder,
                                self.max_output_len)
        embedding_weights_encoder, embedding_weights_decoder = None, None

        if self.embedding_path_encoder is not None:
            self.logger.info('loading encoder embedding from {}'.format(self.embedding_path_encoder))
            embedding = read_embedding(self.embedding_path_encoder, summarizer.embedding_size)
            embedding_weights_encoder = embedding_to_matrix(embedding=embedding,
                                                            token_index=tokenizer_encoder.token_index,
                                                            embedding_dim=summarizer.embedding_size)
            num_unknown_encoder = tokenizer_encoder.vocab_size - len(embedding.keys())
            self.logger.info('unknown vocab encoder embedding: {}'.format(num_unknown_encoder))

        if self.embedding_path_decoder is not None:
            self.logger.info('loading decoder embedding from {}'.format(self.embedding_path_decoder))
            embedding = read_embedding(self.embedding_path_decoder, summarizer.embedding_size)
            embedding_weights_decoder = embedding_to_matrix(embedding=embedding,
                                                            token_index=tokenizer_decoder.token_index,
                                                            embedding_dim=summarizer.embedding_size)
            num_unknown_decoder = tokenizer_decoder.vocab_size - len(embedding.keys())
            self.logger.info('unknown vocab encoder embedding: {}'.format(num_unknown_decoder))

        summarizer.init_model(preprocessor=self.preprocessor,
                              vectorizer=vectorizer,
                              embedding_weights_encoder=embedding_weights_encoder,
                              embedding_weights_decoder=embedding_weights_decoder)

    def _vectorize_data(self,
                        preprocessor: Preprocessor,
                        vectorizer: Vectorizer,
                        bucket_generator: BucketGenerator = None) -> Callable[[Iterable[Tuple[str, str]]],
                                                                              Iterable[Tuple[List[int], List[int]]]]:

        def vectorize(raw_data: Iterable[Tuple[str, str]]):
            data_preprocessed = (preprocessor(d) for d in raw_data)
            data_vectorized = (vectorizer(d) for d in data_preprocessed)
            if bucket_generator is None:
                return data_vectorized
            else:
                return self.bucket_generator(data_vectorized)

        return vectorize

    def _create_tokenizers(self,
                           train_data: Iterable[Tuple[str, str]]
                           ) -> Tuple[KerasTokenizer, KerasTokenizer]:

        self.logger.info('fitting tokenizers...')
        counter_encoder = Counter()
        counter_decoder = Counter()
        train_preprocessed = (self.preprocessor(d) for d in train_data)
        for text_encoder, text_decoder in train_preprocessed:
            counter_encoder.update(text_encoder.split())
            counter_decoder.update(text_decoder.split())
        tokens_encoder = {token_count[0] for token_count in counter_encoder.most_common(self.max_vocab_size_encoder)}
        tokens_decoder = {token_count[0] for token_count in counter_decoder.most_common(self.max_vocab_size_decoder)}
        tokens_encoder.update({self.preprocessor.start_token, self.preprocessor.end_token})
        tokens_decoder.update({self.preprocessor.start_token, self.preprocessor.end_token})
        tokenizer_encoder = KerasTokenizer(oov_token=OOV_TOKEN, lower=False, filters='')
        tokenizer_decoder = KerasTokenizer(oov_token=OOV_TOKEN, lower=False, filters='')
        tokenizer_encoder.fit(sorted(list(tokens_encoder)))
        tokenizer_decoder.fit(sorted(list(tokens_decoder)))
        return tokenizer_encoder, tokenizer_decoder
